'''
Module providing Database Views
'''

from collections.abc import Sequence 
from typing import Any

from playhouse.shortcuts import model_to_dict

from .model import *
from .utils import tables

class Cache(Sequence):
    '''
    Sequence Class caching Data from a BaseModel Table 
    '''
    def __init__(self, table: BaseModel) -> None:
        assert any(table.__name__ == t.__name__ for t in tables), f"Must provide a BaseModel Object; not {type(table)}"

        self.table = table
        self.data = None

    def update(self, value: Any, update_param: str = 'id') -> None:
        '''
        Updates Cache by selecting data given by self.table.<update_param> == value 

        '''
        assert update_param in self.table._meta.sorted_field_names, f"Update Parameter must exist on {self.table.__name__}"
        
        param = getattr(self.table, update_param)
        data = self.table.select().where(param == value)
        self.data = list(data) if data else None

    def is_None(self) -> bool:
        '''
        Check if Cache is Empty
        '''
        return self.data is None

    def __getitem__(self, id: int) -> BaseModel | None:
        '''
        Get item from Cache based on provided id
        '''
        assert type(id) is int, "ID must be an Integer"
        for el in self.data:
            if el.get_id() == id:
                return el
        return None
 
    def __len__(self) -> int:
        '''
        returns Number of Elements Stored in Cache
        '''
        return len(self.data)

    def __contains__(self, data: BaseModel | int) -> bool:
        '''
        Check if provided BaseModel object is stored in Cache
        '''
        assert isinstance(data, (self.table, int)), f"Check could only be made if Object of type {self.table.__name__} or integer id"

        if isinstance(data, BaseModel):
            return data in self.data
        if isinstance(data, int):
            for el in self.data:
                if el.get_id() == data:
                    return True
        return False

    def __iter__(self) -> BaseModel:
        '''
        Generator returning BaseModel Objects stored in Cache
        Note: Like SQL there is no real Order to the Data
        '''
        yield from self.data 

    def __reversed__(self) -> BaseModel:
        '''
        Same as __iter__ but reversed
        '''
        yield from reversed(self.data)
    
    def index(self, value, start=0, stop=None):
        '''
        Not Implemented do to SQL Order Issue
        '''
        return NotImplemented

    def count(self, value):
        '''
        Not Implemented do to assumed Uniqness of Data
        '''
        return NotImplemented

    def to_dict(self, recurse=False) -> list[dict]:
        '''
        Returns a list containing the resolved BaseModel data as dict 
        
        If recurse == True Database trys to recurse all Relationships
        Note: This Operation isn't Cache performed!

        Format:
        [
            {<TableParam>: <Value>, ...},
            ...
        ]
        '''
        assert type(recurse) is bool, "recurse must be bool"
        return [model_to_dict(data, recurse=recurse) for data in self.data]

    def filter(self):
        '''
        WIP
        '''
        pass

if __name__ == "__main__":
    from utils import init_db
    init_db('test.db')
    from pprint import pprint
   
    help(TownCache)
